#include <stdio.h>
#include <stdlib.h>
#include <phidget21.h>

CPhidgetDictionaryHandle dict = 0;
CPhidgetDictionaryListenerHandle robotSetListener = 0;

CPhidgetSpatialHandle spatial;

CPhidgetMotorControlHandle motoControl[4] = {0,0,0,0};
const char *motorLabels[4] = {"FrontLeft","FrontRight","RearLeft","RearRight"};
int motorCount = 0;
int gotSet = 0;
double heading = 0;

int robotSetChangeHandler(CPhidgetDictionaryHandle dict, void *userPtr, const char *key, const char *value, CPhidgetDictionary_keyChangeReason reason)
{
	const char *reasonString;
	char *endPtr;
	double valuenum;

	if(reason != PHIDGET_DICTIONARY_ENTRY_REMOVING)
	{
		if(!strcmp(key, "/robot/setMotorSpeeds"))
		{
			valuenum = strtod(value, &endPtr);
			CPhidgetMotorControl_setVelocity(motoControl[0], 0, valuenum);
			
			valuenum = strtod(endPtr+1, &endPtr);
			CPhidgetMotorControl_setVelocity(motoControl[1], 0, valuenum);
			
			valuenum = strtod(endPtr+1, &endPtr);
			CPhidgetMotorControl_setVelocity(motoControl[2], 0, valuenum);
			
			valuenum = strtod(endPtr+1, &endPtr);
			CPhidgetMotorControl_setVelocity(motoControl[3], 0, valuenum);
			
			CPhidgetDictionary_addKey(dict, "/robot/stat/MotorSpeeds", value, PFALSE);
			gotSet = 0;
		}
		else if(!strcmp(key, "/robot/setMotorAccels"))
		{
			valuenum = strtod(value, NULL);
			
			//printf("Setting Acceleration to: %lf\n", valuenum);
			
			CPhidgetMotorControl_setAcceleration(motoControl[0], 0, valuenum);
			CPhidgetMotorControl_setAcceleration(motoControl[1], 0, valuenum);
			CPhidgetMotorControl_setAcceleration(motoControl[2], 0, valuenum);
			CPhidgetMotorControl_setAcceleration(motoControl[3], 0, valuenum);
			
			CPhidgetDictionary_addKey(dict, "/robot/stat/MotorAccels", value, PFALSE);
			gotSet = 0;
		}
	}
	
	return 0;
}

int ServerConnectedHandler(CPhidgetHandle phid, void *userPtr)
{
	char val[1024];
	
	snprintf(val, 1024, "%d", motorCount);
	CPhidgetDictionary_addKey(dict, "/robot/stat/attachCount", val, PFALSE);
	CPhidgetDictionary_addKey(dict, "/robot/stat/running", "true", PFALSE);
	
	printf("Server Connection Established.\n");
	CPhidgetDictionary_set_OnKeyChange_Handler(dict, &robotSetListener, "^/robot/setMotor.*$", robotSetChangeHandler, NULL);
	return 0;
}

int ServerDisconnectedHandler(CPhidgetHandle phid, void *userPtr)
{
	printf("Server Disconnected!\n");
	CPhidgetDictionary_remove_OnKeyChange_Handler(robotSetListener);
	robotSetListener = 0;
	return 0;
}

int spatialAttach(CPhidgetHandle MC, void *userPtr)
{
	int serialNo;
	const char *name;

	CPhidget_getDeviceName (MC, &name);
	CPhidget_getSerialNumber(MC, &serialNo);
	printf("%s %10d attached!\n", name, serialNo);
	
	CPhidgetSpatial_setDataRate(spatial, 64);
	CPhidgetSpatial_zeroGyro(spatial);
}

int AttachHandler(CPhidgetHandle MC, void *userPtr)
{
	int serialNo;
	const char *name;
	char val[1024];
	
	motorCount++;
	
	snprintf(val, 1024, "%d", motorCount);
	CPhidgetDictionary_addKey(dict, "/robot/stat/attachCount", val, PFALSE);

	CPhidget_getDeviceName (MC, &name);
	CPhidget_getSerialNumber(MC, &serialNo);
	printf("%s %10d attached! (%s)\n", name, serialNo, motorLabels[(int)userPtr]);

	CPhidgetMotorControl_setAcceleration((CPhidgetMotorControlHandle)MC, 0, 250);
	CPhidgetMotorControl_setBraking((CPhidgetMotorControlHandle)MC, 0, 100);
	return 0;
}

int DetachHandler(CPhidgetHandle MC, void *userPtr)
{
	int serialNo;
	const char *name;
	char val[1024];
	
	motorCount--;
	
	snprintf(val, 1024, "%d", motorCount);
	CPhidgetDictionary_addKey(dict, "/robot/stat/attachCount", val, PFALSE);

	CPhidget_getDeviceName (MC, &name);
	CPhidget_getSerialNumber(MC, &serialNo);
	printf("%s %10d detached! (%s)\n", name, serialNo, motorLabels[(int)userPtr]);

	return 0;
}

int VelocityChangeHandler(CPhidgetMotorControlHandle MC, void *usrPtr, int Index, double Value)
{
	char val[1024], key[1024];
	
	snprintf(key, 1024, "%lf", Value);
	snprintf(val, 1024, "/robot/stat/%sSpeed", motorLabels[(int)usrPtr]);
	
	//printf("Motor %d > Current Speed: %f\n", Index, Value);
	CPhidgetDictionary_addKey(dict, val, key, PFALSE);
	return 0;
}

int SpatialDataHandler(CPhidgetSpatialHandle spatial, void *userptr, CPhidgetSpatial_SpatialEventDataHandle *data, int count)
{
	char val[1024];
	static int lastTimeValid = 0;
	static CPhidget_Timestamp lastTime;
	
	if(lastTimeValid)
	{
		double timeChange = data[0]->timestamp.seconds - lastTime.seconds + ((data[0]->timestamp.microseconds - lastTime.microseconds) / 1000000.0);
		heading += (data[0]->angularRate[2] * timeChange);
		
		snprintf(val, 1024, "%0.1lf", heading);
		
		//printf("Motor %d > Current Speed: %f\n", Index, Value);
		CPhidgetDictionary_addKey(dict, "/robot/stat/heading", val, PFALSE);
	}
	
	lastTime = data[0]->timestamp;
	lastTimeValid = 1;
	

	return 0;
}

int main(int argc, char* argv[])
{
	int result, serverStatus;
	int i;

	result = CPhidgetDictionary_create(&dict);

	result = CPhidget_set_OnServerConnect_Handler((CPhidgetHandle)dict, ServerConnectedHandler, NULL);
	result = CPhidget_set_OnServerDisconnect_Handler((CPhidgetHandle)dict, ServerDisconnectedHandler, NULL);

	result = CPhidgetDictionary_openRemoteIP(dict, "localhost", 5001, NULL);

	do {
		result = CPhidgetDictionary_getServerStatus(dict, &serverStatus);
	} while(serverStatus = PHIDGET_NOTATTACHED);

	printf("Attached to server\n");

	for(i=0;i<4;i++)
	{
		CPhidgetMotorControl_create(&motoControl[i]);

		CPhidget_set_OnAttach_Handler((CPhidgetHandle)motoControl[i], AttachHandler, (void *)i);
		CPhidget_set_OnDetach_Handler((CPhidgetHandle)motoControl[i], DetachHandler, (void *)i);
		
		//CPhidgetMotorControl_set_OnVelocityChange_Handler (motoControl[i], VelocityChangeHandler, (void *)i);

		CPhidget_openLabel((CPhidgetHandle)motoControl[i], motorLabels[i]);
	}
	
	CPhidgetSpatial_create(&spatial);
	CPhidget_set_OnAttach_Handler((CPhidgetHandle)spatial, spatialAttach, NULL);
	CPhidgetSpatial_set_OnSpatialData_Handler(spatial, SpatialDataHandler, NULL);
	CPhidget_open((CPhidgetHandle)spatial, -1);
	
	while(1)
	{
		usleep(100000); //100 ms
		gotSet++;
		
		//We haven't had a set in > 500ms!! Stop motors
		if(gotSet >= 5)
		{
			CPhidgetMotorControl_setVelocity(motoControl[0], 0, 0);
			CPhidgetMotorControl_setVelocity(motoControl[1], 0, 0);
			CPhidgetMotorControl_setVelocity(motoControl[2], 0, 0);
			CPhidgetMotorControl_setVelocity(motoControl[3], 0, 0);
		}
	}

	result = CPhidgetDictionary_close(dict);
	result = CPhidgetDictionary_delete(dict);
	return 0;
}
